import torch
import torch.optim as optim
import torch.nn as nn
import argparse
import os
import random
from torch.autograd import Variable
from torch.utils.data import DataLoader
import utils3_48 as utils
import itertools
import progressbar
import numpy as np

import pickle
from torch.distributions.normal import Normal

# --------- training funtions ------------------------------------
from torch.distributions.normal import Normal

def train(x):
    ########## input: x: List of length opt.n_past + opt.n_future time steps. Each element of the list is Batch Size x # Channels x Height x Width ##########

    # initialize the hidden state of the LSTM modules.
    frame_predictor.hidden = frame_predictor.init_hidden()
    posterior.hidden = posterior.init_hidden()
    prior.hidden = prior.init_hidden()

    # Initialize the loss terms
    disc_loss, gen_loss, real_loss, syn_loss, wt_gen_loss = 0, 0, 0, 0, 0
    syn_lstm_loss, real_lstm_loss, gen_lstm_loss = 0, 0, 0
    pred_vid_frame_buff, real_vid_frame_buff = [], []
    wt_mse, mse, reg_sig, prob_reg = 0, 0, 0, 0
    kld, gt_kld = 0, 0 #*, 1e-7
    prob_vals, sig_vals = [], []

    # Iterate over each frame one step at a time
    for i in range(1, opt.n_past+opt.n_future):
        h = encoder(x[i-1]) # Push the frame x_{i-1} into the encoder
        if opt.last_frame_skip or i <= opt.n_past:   
            h, skip = h
        else:
            h = h[0]
        h_target = encoder(x[i])[0] # Push the frame x_{i} into the encoder

        # Obtain latent space statistics
        z_t, mu, logvar = posterior(h_target) 
        _, mu_p, logvar_p = prior(h)

        # Obtain the decoded output
        h_pred = frame_predictor(torch.cat([h, z_t], 1))
        x_pred = decoder([h_pred, skip])

        # Obtain the variance of the MSE, and the regularizations over it - use truncated normal distribution
        sigma_prob, mu_op, sig_op, reg_sig = sig_trans(torch.div(torch.exp(logvar_p), torch.exp(logvar_p).sum(dim=1, keepdim=True)).detach() + opt.eps)  

        # Compute the losses
        wt_mse += ((1. - sigma_prob).squeeze() * torch.flatten(mse_criterion(x_pred, x[i]), start_dim=1).mean(dim=1) ).mean() #* Note we predict (1 - s_t)
        kld += kl_criterion(mu, logvar, mu_p, logvar_p) # KL-Divergence loss in latent space
        
        # ---------------- Hyper-prior Computation ----------------------
        # Initialize the statistics of the hyperprior
        mu_gt = torch.zeros_like(mu_op) + opt.dist_mu
        sig_gt = torch.zeros_like(sig_op) + opt.dist_sig

        gt_kld += kl_criterion(mu_op, sig_op, mu_gt, sig_gt)

        # --------------- Discriminator to judge if a frame is real or synthetic ----------------------------
        # Encode the synthetic frame
        x_hat_enc, _ = disc_encoder(x_pred) 

        # Generate labels for discriminator
        l_z = torch.LongTensor(x[i].size(0)); l_z.fill_(0); l_z = l_z.cuda();
        l_o = torch.LongTensor(x[i].size(0)); l_o.fill_(1); l_o = l_o.cuda();

        # Get the classification labels and loss for the real data
        real_pred = disc(h_target.detach())
        real_loss += classif_criterion(real_pred, l_o).mean()

        # Get the classification labels and loss for the synthetic data
        syn_pred = disc(x_hat_enc.detach())
        syn_loss += classif_criterion(syn_pred, l_z).mean()

        # Get the classification labels and loss for the synthetic data for training the generator
        gen_pred = disc(x_hat_enc)

        # Obtain the gen_loss term for the tensor
        wt_gen_loss += classif_criterion(gen_pred, l_o).mean() 

        # ---------------- LSTM based discriminator -------------------------
        # Forward pass through encoder
        pred_vid_frame_buff.append(x_hat_enc) # Append the current predicted frame to the end
        real_vid_frame_buff.append(h_target) # Append the target video frame to the end

        if len(pred_vid_frame_buff) > opt.disc_buffer: # The buffer is full- then pop
            pred_vid_frame_buff, real_vid_frame_buff = pred_vid_frame_buff[1:], real_vid_frame_buff[1:]
        elif len(pred_vid_frame_buff) < opt.disc_buffer: # The buffer is full: # We don't compute gradients until the buffer is full capacity
            continue

        # Forward pass through the discriminator - with synthetic samples for discriminator
        disc_lstm.hidden = disc_lstm.init_hidden()
        # Iterate through the buffer
        for j in range(len(pred_vid_frame_buff)):
            if j == (opt.disc_buffer - 1): # Executing the final frame of the buffer
                pred_prob_vec = disc_lstm(pred_vid_frame_buff[j].detach())
            else:
                disc_lstm(pred_vid_frame_buff[j].detach())
        # Generate labels for discriminator
        l_z = torch.LongTensor(x[i].size(0)); l_z.fill_(0); l_z = l_z.cuda();
        l_o = torch.LongTensor(x[i].size(0)); l_o.fill_(1); l_o = l_o.cuda();
        # Compute the classification loss for the final frame of the buffer
        syn_lstm_loss += classif_criterion(pred_prob_vec, Variable(l_z)).mean() 

        # Forward pass through the discriminator - with real samples 
        disc_lstm.hidden = disc_lstm.init_hidden()
        # Iterate through the buffer
        for j in range(len(real_vid_frame_buff)):
            if j == (opt.disc_buffer - 1): # Executing the final frame of the buffer
                real_prob_vec = disc_lstm(real_vid_frame_buff[j].detach())
            else:
                disc_lstm(real_vid_frame_buff[j].detach())
        # Compute the loss for the real samples
        real_lstm_loss += classif_criterion(real_prob_vec, Variable(l_o)).mean() 

        # Forward pass through the discriminator - with synthetic samples for the generator
        disc_lstm.hidden = disc_lstm.init_hidden()
        # Iterate through the buffer
        for j in range(len(pred_vid_frame_buff)):
            if j == (opt.disc_buffer - 1): # Executing the final frame of the buffer
                pred_prob_vec_gen = disc_lstm(pred_vid_frame_buff[j])
            else:
                disc_lstm(pred_vid_frame_buff[j].detach())
        # Compute the generator loss: Minimize -log D(G(z))
        gen_lstm_loss += classif_criterion(pred_prob_vec_gen, Variable(l_o)).mean() 

    return real_loss, real_lstm_loss, syn_loss, syn_lstm_loss, wt_mse, reg_sig, wt_gen_loss, gen_lstm_loss, kld, gt_kld  

